;
; GRDP
;
; Copyright(c) LADsoft
;
; David Lindauer, gclind01@starbase.spd.louisville.edu
;
;
; FP.ASM
;
; Function: Handle FP status commands
;
	;MASM MODE
	.MODEL SMALL
	.386

	public	floatcheck, fpcommand

include  prints.ase 
include  input.ase 
include  mtrap.ase 
include  breaks.ase 
include  loader.ase
include  dos.ase
include exec.ase

	.data
ten	dw	10
work	dw	?
fpflag	db	?
floatstat dw	47 dup (0)
enames	db	"IDZOUP"


	.code
precmsg	dw	offset _text:prsing,offset _text:reserved
	dw	offset _text:prdbl,offset _text:prxt
roundmsg dw	offset _text:rdnear,offset _text:rdminus
	dw	offset _text:rdplus,offset _text:rdzer
prsing	db	"Single",0
prdbl	db	"Double",0
prxt	db	"Extended",0
reserved db	"Reserved",0
rdnear	db	"Nearest",0
rdzer	db	"Zero",0
rdplus	db	"Plus Infinity",0
rdminus db	"Minus Infinity",0

;
; check for floating point unit
;
floatcheck	proc
		mov	[fpflag],1
		mov	[work],4567h
		fninit
		fstsw 	[work]
		cmp	[work],0
		jz	hasfp
		mov	[fpflag],0
hasfp:
		ret
floatcheck	endp
;
; fp commands
;
fpcommand	proc
	test	[fpflag],1
	jnz	hasfloat
	Msg	<13,10,"No FP unit">
	clc
	ret
hasfloat:
	fnsave	[floatstat]
	fwait
	call	WadeSpace
	jz	dumpstack
	cmp	al,'s'
	jnz	flerr
	inc	si
	call	WadeSpace
	jnz	flerr
	frstor [floatstat]
	fwait
	Msg	<13,10,"Masked exceptions: ">
	mov	dx,word ptr [floatstat]
	call	except
	Msg	<13,10,"Active exceptions: ">
	mov	dx,word ptr [floatstat+2]
	call	except
	Msg	<13,10,"Precision:         ">
	movzx	bx,byte ptr [floatstat+1]
	and	bl,3
	shl	bl,1
	mov	bx,[bx+precmsg]
	call	olMessage
	Msg	<13,10,"Rounding:          ">
	movzx	bx,byte ptr [floatstat+1]
	and	bl,0ch
	shr	bl,1
	mov	bx,[bx+roundmsg]
	call	olMessage
	Msg	<13,10,"FPU Status flags:  ">
	test	word ptr [floatstat+2],100h
	jz	notc0
	Msg	"C0 "
notc0:
	test	word ptr [floatstat+2],200h
	jz	notc1
	Msg	"C1 "
notc1:
	test	word ptr [floatstat+2],400h
	jz	notc2
	Msg	"C2 "
notc2:
	test	word ptr [floatstat+2],4000h
	jz	notc3
	Msg	"C3 "
notc3:
	clc
	ret
dumpstack:
	mov	ax,word ptr [floatstat+2]
	shr	ax,11
	and	ax,7
	sub	ax,7
	neg	ax
	mul	[ten]
	mov	bx,offset dgroup:floatstat+14
	add	bx,ax
	mov	cx,8
dsl:
	push	bx
	push	cx
	call	dumpval
	pop	cx
	pop	bx
	sub	bx,10
	loop	dsl
	frstor [floatstat]
	fwait
	clc
	ret
flerr:
	stc
	ret
fpcommand	endp
except	PROC
	mov	si,offset dgroup:enames
	mov	cx,6
exlp:
	shr	dx,1
	lodsb
	jnc	nhr
	push	dx
   	mov	dl,al
	call	putchar
	mov	dl,'E'
	call	putchar
	call	printspace
	pop	dx
nhr:
	loop	exlp
	ret
except	ENDP
;
; dump the value of a stack register
;
dumpval	PROC
	call	crlf
	dec	cl
	mov	dl,cl
	add	dl,'0'
	call	putchar
	mov	ax,word ptr [floatstat+2]
	shr	ax,11
	and	al,7
	cmp	al,cl
	jnz	nottop
       	mov	dl,'*'
	call	putchar
	jmp	join
nottop:
	call	printspace
join:
	mov	dl,')'
	call	putchar
	call	printspace
	call	printspace
	call	printspace
	mov	ax,word ptr [floatstat+4]
	shr	ax,cl
	shr	ax,cl
	and	al,3
	cmp	al,3
	jnz	notempty
	Msg	"<Empty>"
	clc
	ret

notempty:
	cmp	al,2
	jnz	notnan
isnan:
	Msg	"<Nan>"
	clc
	ret
notnan:
	cmp	al,1
	jnz	notzero
	mov	dl,'0'
	call	putchar
	clc
	ret
notzero:
	fld	tbyte ptr [bx]
	call	PrintFloating
	fcomp
	clc
	ret
dumpval	ENDP
;
; actual FP print routines
;
PrintFloating	PROC
	call	fextract
	or	ax,ax
	jz	notdenorm
	Msg	"<Denorm>"
	ret
notdenorm:
	or	dx,dx
	jns	noneg
	call	putneg
noneg:
	cmp	bx,8		; getter be less than 16!!!!
	jg	highexp
	cmp	bx,-4
	jle	lowexp
	or	bx,bx
	jge	gte0
;
; print for less than one but not exponential
;
	push	bx
	call	putzer
	call	putper
	pop	cx
	neg	cx
	dec	cx
	jcxz	nozr
	inc	cx
zrl:
	call	putzer
	loop	zrl
nozr:
	mov	cx,16
nl2:
	call	fnd
	call	putdig
	loop	nl2
	ret
;
; print for greater than or equal to one but not exponential
;
gte0:
	push	bx
	mov	cx,bx
	call	fnd
	call	putdig
	jcxz	gte0nl
gtel1:
	call	fnd
	call	putdig
	loop	gtel1
gte0nl:
	call	putper
	pop	cx
	sub	cx,16
	neg	cx
gtel2:
	call	fnd
	call	putdig
	loop	gtel2
	ret
;
; hack to justify negative exponents
;
lowexp:
	dec	bx
	call	fnd
;
; exponential print
;
highexp:
	push	bx
	call	fnd
	call	putdig
	call	putper
	mov	cx,16
hel1:
	call	fnd
	call	putdig
	loop	hel1
	mov	dl,'E'
	call	putchar
	pop	ax
	or	ax,ax
	jns	nonegx
	push	ax
	call	putneg
	pop	ax
	neg	ax
nonegx:
	sub	cx,cx
elp1:
	sub	dx,dx
	div	[ten]
	push	dx
	inc	cx
	or	ax,ax
	jnz	elp1
elp2:
	pop	ax
	call	putdig
	loop	elp2
	ret
putdig:
	mov	dl,al
	add	dl,'0'
	call	putchar
       	ret
putper:
	mov	dl,'.'
	call	putchar
	ret
putneg:
	mov	dl,'-'
	call	putchar
	ret
putzer:
	mov	dl,'0'
	call	putchar
	ret
PrintFloating	ENDP
;
; get exponent and mantissa and sign
;
; enter:
;   st(0) = value
;
; returns:
;   ax = 0	; ok val
;   
;     st(0) = mantissa
;     bx = exp
;     dx = sign   ( +-1)
;
;   ax = -1	; infinity or dnormal
;
fextract 	PROC
	enter	6,0
	fstcw	[bp-2]
	mov	ax,[bp-2]
	or	ah,0ch		; round toward zero
	mov	[bp-4],ax
	fldcw	[bp-4]

	ftst
	fnstsw	ax
	fnclex			; just in case ... 
	sahf
	mov	eax,-1   	; first check for out of range
	jp	fxx
	jnz	dosign		; now check for zero
	sub	bx,bx
	sub	dx,dx
	jmp	fxx

dosign:
	mov	dx,1		; nonzero, finally get sign
	jnc	pos
	fabs       		; we will work with positive nums hereafter
	mov	dx,-1
pos:
	fldlg2       		; log to base 10
	fxch
	fyl2x
	
	fld	st(0)		; get int part
	frndint
	fist	word ptr [bp-6]
	pop	bx

	fsubp			; fraction
	fldl2t			; convert back to base 2
	fmulp

	fld	st(0)		; lovely exponentiation
	frndint
	fxch
	fld	st(1)
	fsubp
	f2xm1			;
	fld1
	faddp
	fscale
	fxch
	fcomp
	sub	eax,eax
fxx:
	fnclex
	fldcw	[bp-2]
	fwait
	leave
	ret
fextract	ENDP
;
; get next digit from mantissa
;
; enter:
;   mantissa from fextract on stack
;
; exit:
;   eax = next digit (base 10)
;   stack = new mantissa
;
fnd	PROC
	ENTER	6,0
	fstcw	[bp-2]
	mov	ax,[bp-2]
	or	ah,0ch		; round toward zero
	mov	[bp-4],ax
	fldcw	[bp-4]

	fld	st(0)		; next digit
	frndint
	fist	word ptr [bp-6]

	fsubp			; new mantissa
	fimul	word ptr [ten]

	fnclex
	fldcw	[bp-2]
	fwait
	pop	ax		; cute trick to get result :)
	leave
	ret
fnd	ENDP
end