;
; GRDB
;
; Copyright(c) LADsoft
;
; David Lindauer, gclind01@starbase.spd.louisville.edu
;
;
; Input.asm
;
; Function: Handle input
;
;   Handles numbers
;   Handles segments
;   Handles trapping page faults
;   Handles command input
;
	;MASM MODE
	.model small
	.386

include  dispatch.ase 
include  prints.ase 
include  mtrap.ase 
include  regs.ase 
include  dump.ase 
include  entry.ase 
include  exec.ase 
include  breaks.ase 
include  dis.ase 
include  inout.ase
include  options.ase
include  history.ase

	extrn _exit : PROC,domath : PROC, drive : PROC, fill : PROC
	extrn logging : PROC, help : PROC
	extrn DoFileRead : PROC, DoFileWrite : PROC, search : PROC
	extrn fpcommand : PROC

	PUBLIC	qerr, ReadNumber, ReadAddress, Inputhandler
	PUBLIC	wadeSpace, GetInputLine, WadeSpaceOnly, inputbuffer
IBSIZE = 80

	.data
inputbuffer db	IBSIZE DUP (?)		; Input buffer
Commands db	"bdefghilnopqrstuwx@?"		; List of commands
comlen	= $ - Commands			; Length of list
inquote db	0

	.code
;
; Print an error if command wrong
;
qerr	PROC	
	call	crlf
	sub	si,offset dgroup:inputbuffer-2; Calculate error pos
	mov	cx,si			;
	jcxz	short qestart		;
	dec	cx			;
	jcxz	short qestart		;
qelp:					
	call	printspace              ; Space over to error pos
     	loop	qelp
qestart:
	mov	dl,'^'			; Display error
	call	PutChar
	stc				; Did an error
	ret	
qerr	ENDP	
;
; Read in a number
;
ReadNumber	PROC	
	push	ebx
	push	cx
	push	dx
	call	ReadReg
	clc
	jz	gotnum
	sub	ebx,ebx			; Number = 0
	sub	cx,cx			; digits = 0
	mov	al,[si]
	cmp	al,"'"
	jz	getstring
	cmp	al,'"'
	jz	getstring
rnlp:
	lodsb				; Get char & convert to uppercase
	cmp	al,60h
	jc	notlower
	and	al,NOT 20h
notlower:
	sub	al,'0'    		; Convert to binary
	jc	short rn_done		; < '0' is an error
	cmp	al,10			; See if is a digit
	jc	short gotdigit		; Yes, got it
	sub	al,7			; Convert letters to binary
	cmp	al,16			; Make sure is < 'G'
	jnc	short rn_done		; Quit if not
	cmp	al,10			; MAke sure not < 'A'
	jc	short rn_done
gotdigit:
	shl	ebx,4			; It is a hex digit, add in
	or	bl,al			;
	inc	cx			; Set flag to indicate we got digits
	jmp	rnlp
rn_done:
	dec	si			; Point at first non-digit
rm_done2:
	mov	eax,ebx
	test	cl,-1			; See if got any
	jnz	gotnum			;
	stc				; No, error
gotnum:
	pop	dx
	pop	cx
	pop	ebx
	ret	
getstring:
	inc si
getstringl:
	lodsb
	cmp	al,13
	jz	rn_done
	cmp	al,"'"
	jz	rm_done2
	cmp	al,'"'
	jz	rm_done2
	inc	cl
	shl	ebx,8
	mov	bl,al
	jmp	getstringl
ReadNumber	ENDP	
;
; Read an address, composed of a number and a possible selector
;
ReadAddress	PROC	
	sub	edx,edx
	call	ReadReg
	jnc	gotseg
	mov	ebx,eax
	jz	gotaddr
readseg:
	call	ReadNumber              ; Read a number
	jc	short raerr		; Quit if error
	mov	ebx,eax
	call	WadeSpace
	jz	gotaddr
	cmp	al,':'
	jnz	gotaddr
	inc	si
	mov	dx,bx			; Number to EBX
	or	edx,10000H
	jmp	readofs
gotseg:
	sub	ebx,ebx
	mov	dx,ax
	or	edx,10000H
	call	WadeSpace
	jz	gotaddr
	cmp	al,':'
	jnz	gotaddr
	inc	si
readofs:
	call	WadeSpace
	jz	gotaddr
	call	ReadNumber		; Read in offset
	jc	short raerr		; Quit if error
	mov	ebx,eax			;
gotaddr:
	clc				; OK, exit
	ret
raerr:
	stc				; Error on number input
	ret
ReadAddress	ENDP	
;
; Get an input line
;
GetInputLine	PROC	
	mov	[inquote],0
	mov	di,offset dgroup:InputBuffer	; Get input buffer
	mov	si,di			; Return buffer pointer
	mov	cx,IBSIZE		; Size of buffer
moreinput:
	call	GetKey
	or	ax,ax    		; ignore ctrl-break
	jz	moreinput
	call	CheckHistory
	jc	MoreInput
	cmp	al,8			; Is delete or rubout?
	jz	short bkspc		; Yes - go do it
	cmp	al,7fh			;
	jz	short bkspc		; yes - go do it
	cmp	al,'"'
	jz	doquote
	cmp	al,"'"
	jnz	dochar
doquote:
	xor	[inquote],1
dochar:
	push	ax
	test	[inquote],1
	jnz	nolc
	cmp	al,'A'
	jc	nolc
	cmp	al,'Z'
	ja	nolc
	or	al,20h
nolc:
	stosb
	pop	ax
	cmp	al,13			; Is CR
	jz	short endinput		; Yes, return
	mov	dl,al			; Echo character
	call	PutChar
	loop	moreinput		; Loop till buffer full
endinput:
	call	EnterHistory
	ret
bkspc:
	cmp	di,offset dgroup:InputBuffer	; Quit if nothing in buffer
	jz	moreinput		; And get more input
	mov	dl,8			; Erase last echoed char
	call	PutChar
	mov	dl,' '			;
	call	PutChar
	mov	dl,8			; Reset pointer
	call	PutChar
	dec	di			; Point at last char
	jmp	moreinput		; Get more input
GetInputLine	ENDP	
;
; Wade pasth spaces
;
WadeSpace	PROC	
	lodsb				; Get char
	cmp	al,' '			; if ' ' or ',' go again
	jz	short wadeSpace		;
	cmp	al,','			;
	jz	short WadeSpace		;
	dec	si			; Point at last space char
	cmp	al,13
	ret
WadeSpace	ENDP	
;
; Wade through spaces only
;
WadeSpaceOnly	PROC	
	lodsb			; Get a char
	cmp	al,' '		; Is space
	jz	WadeSpaceOnly	; Loop if so
	dec	si		; Else point at char
	cmp	al,13
	ret
WadeSpaceOnly	ENDP	
;
; Main Input routine
;
InputHandler	PROC
	call	LoadHistory
	Msg	<13,10,"* ">		; MONITOR prompt
	call	GetInputLine		; Get an input line
	call	WadeSpace		; Wade through spaces
	jz	InputHandler		;
	inc	si			; Point at first non-space char
	mov	di,offset dgroup:commands	; Get command list
	mov	cx,comlen		; Length of list
	repne	scasb			;
	jnz	ierr			; Error if not in list
	mov	ax,comlen-1		; Calculate position
	sub	ax,cx			;
	push	0			; Command arg = 0
	call	TableDispatch		; Dispatch command
	dw	comlen-1
	dw	breaks
	dw	Dump
	dw	entry
	dw	fill
	dw	go
	dw	domath
	dw	doin
	dw	DoFileRead
	dw	fpcommand
	dw	doout
	dw	proceed
	dw	_exit
	dw	ModifyRegisters
	dw	search
	dw	trap
	dw	diss
	dw	DoFileWrite
	dw	drive
	dw	logging
	dw	help
	jnc	InputHandler		; Get more input if no err
ierr:
	call	qerr			; Display error
	jmp	InputHandler		; Get more input
InputHandler	ENDP
END